import asyncio
import logging
import shutil
from collections import OrderedDict
from typing import Any, Coroutine, Dict, List, Text

from sanic import Blueprint, response
from sanic.request import Request

import rasax.community.version
from rasa.core.domain import InvalidDomain
from rasax.community import config
from rasax.community.api.decorators import (
    inject_rasa_x_user,
    platform_scoped,
    validate_schema,
)
from rasax.community.services.domain_service import DomainService
from rasax.community.services.feature_service import FeatureService
from rasax.community.services.settings_service import SettingsService, ProjectException
from rasax.community.services.stack_service import StackService
from rasax.community.services.user_service import (
    UserException,
    UserService,
    MismatchedPasswordsException,
)
from rasax.community.utils import error, dump_yaml, convert_bytes_to_string, default_arg

logger = logging.getLogger(__name__)


def collect_status_calls(service_dict: Dict[Text, StackService]) -> List[Coroutine]:
    """Returns a list of status() Coroutines."""

    return [service.status() for service in service_dict.values()]


def _rasa_services(request: Request) -> Dict[Text, StackService]:
    settings_service = SettingsService(request["db_session"])
    return settings_service.stack_services()


def _domain_service(request: Request) -> DomainService:
    return DomainService(request["db_session"])


async def collect_stack_results(
    stack_services: Dict[Text, StackService], statuses: OrderedDict
) -> Dict[Text, Any]:
    """Creates status result dictionary for stack services."""

    stack_result = dict()

    for name, _status in zip(stack_services.keys(), statuses):
        if isinstance(_status, dict) and "version" in _status:
            _result = _status.copy()
            _result["status"] = 200
        else:
            _result = {"status": 500, "message": _status}
        stack_result[name] = _result

    return stack_result


def blueprint() -> Blueprint:
    project_endpoints = Blueprint("project_endpoints")

    @project_endpoints.route("/health", methods=["GET", "HEAD"])
    async def status(request):
        stack_services = _rasa_services(request)
        stack_status_calls = collect_status_calls(stack_services)
        statuses = await asyncio.gather(*stack_status_calls, return_exceptions=True)
        stack_result = await collect_stack_results(stack_services, statuses)

        return response.json(stack_result)

    @project_endpoints.route("/version", methods=["GET", "HEAD"])
    async def version(request):
        rasa_services = _rasa_services(request)
        stack_status_calls = collect_status_calls(rasa_services)

        statuses = await asyncio.gather(*stack_status_calls, return_exceptions=True)
        rasa_versions = dict()
        for name, _status in zip(rasa_services.keys(), statuses):
            if isinstance(_status, dict) and "version" in _status:
                _result = _status.get("version")
            else:
                _result = {"status": 500, "message": _status}
            rasa_versions[name] = _result

        result = {"rasa": rasa_versions, "rasa-x": rasax.community.__version__}
        return response.json(result)

    @project_endpoints.route("/user", methods=["GET", "HEAD"])
    @platform_scoped("user.get")
    @inject_rasa_x_user()
    async def profile(request, user=None):
        user_service = UserService(request["db_session"])
        return response.json(
            user_service.fetch_user(user["username"], return_api_token=True)
        )

    @project_endpoints.route("/user", methods=["PATCH"])
    @platform_scoped("user.update")
    @inject_rasa_x_user()
    @validate_schema("username")
    async def update_username(request, user=None):
        rjs = request.json

        try:
            user_service = UserService(request["db_session"])
            user_profile = user_service.update_saml_username(
                user["name_id"], rjs["username"]
            )

        except UserException as e:
            return error(
                404,
                "UserException",
                "Could not assign username {} to name_id {}"
                "".format(rjs["username"], user["name_id"]),
                details=e,
            )

        return response.json(user_profile)

    @project_endpoints.route("/users", methods=["GET", "HEAD"])
    @platform_scoped("users.list")
    async def list_users(request):
        user_service = UserService(request["db_session"])
        username_query = default_arg(request, "username", None)
        role_query = default_arg(request, "role", None)
        users = user_service.fetch_all_users(
            config.team_name, username_query, role_query
        )
        if not users:
            return error(404, "NoUsersFound", "No users found")

        profiles = [user_service.fetch_user(u["username"]) for u in users]

        return response.json(profiles, headers={"X-Total-Count": len(profiles)})

    @project_endpoints.route("/users", methods=["POST"])
    @platform_scoped("users.create")
    @validate_schema("user")
    async def create_user(request):
        rjs = request.json
        user_service = UserService(request["db_session"])
        team = rjs.get("team") or config.team_name
        user_service.create_user(
            username=rjs["username"],
            raw_password=rjs["password"],
            team=team,
            roles=rjs["roles"],
        )

        return response.json(rjs, 201)

    @project_endpoints.route("/users/<username>", methods=["DELETE"])
    @platform_scoped("users.delete")
    @inject_rasa_x_user()
    async def delete_user(request, username, user=None):
        user_service = UserService(request["db_session"])

        try:
            deleted = user_service.delete_user(
                username, requesting_user=user["username"]
            )
            return response.json(deleted)
        except UserException as e:
            return error(404, "UserDeletionError", e)

    @project_endpoints.route("/user/password", methods=["POST"])
    @platform_scoped("user.password.update")
    @validate_schema("change_password")
    async def change_password(request):
        rjs = request.json
        user_service = UserService(request["db_session"])

        try:
            user = user_service.change_password(rjs)
            if user is None:
                return error(404, "UserNotFound", "user not found")
            return response.json(user)
        except MismatchedPasswordsException:
            return error(403, "WrongPassword", "wrong password")

    @project_endpoints.route("/projects/<project_id>", methods=["POST"])
    @platform_scoped("projects.create")
    @inject_rasa_x_user()
    async def create_project(request, project_id, user=None):
        settings_service = SettingsService(request["db_session"])

        try:
            project = settings_service.init_project(user["team"], project_id)
        except ProjectException as e:
            return error(404, "ProjectCreationError", e)

        user_service = UserService(request["db_session"])
        user_service.assign_project_to_user(user, project_id)

        return response.json(project)

    # no authentication because features may be needed
    # before a user is authenticated
    @project_endpoints.route("/features", methods=["GET", "HEAD"])
    async def features(request):
        feature_service = FeatureService(request["db_session"])
        return response.json(feature_service.features())

    @project_endpoints.route("/features", methods=["POST"])
    @platform_scoped("features.create", allow_api_token=True)
    @validate_schema("feature")
    async def set_feature(request):
        rjs = request.json
        feature_service = FeatureService(request["db_session"])
        feature_service.set_feature(rjs)
        return response.json(rjs)

    @project_endpoints.route("/logs")
    @platform_scoped("logs.list", allow_api_token=True)
    async def logs(_):
        try:
            shutil.make_archive("/tmp/logs", "zip", "/logs")
            return await response.file("/tmp/logs.zip")
        except Exception as e:
            logger.debug(e)
            return error(500, "error collecting logs", e)

    @project_endpoints.route("/environments", methods=["GET", "HEAD"])
    @platform_scoped("environments.list", allow_api_token=True)
    @inject_rasa_x_user()
    async def get_environment_config(request, user=None):
        settings_service = SettingsService(request["db_session"])
        environments = settings_service.get_environments_config(config.project_name)

        if not environments:
            return error(
                400,
                "EnvironmentSettingsNotFound",
                "could not find environment settings",
            )

        return response.json(
            {"environments": dump_yaml(environments.get("environments"))}
        )

    @project_endpoints.route("/chatToken", methods=["GET", "HEAD"])
    @platform_scoped("chatToken.get", allow_rasa_x_token=True)
    async def get_chat_token(request):
        domain_service = _domain_service(request)
        return response.json(domain_service.get_token())

    @project_endpoints.route("/chatToken", methods=["PUT"])
    @platform_scoped("chatToken.update", allow_api_token=True)
    @validate_schema("update_token")
    async def update_chat_token(request):
        domain_service = _domain_service(request)
        domain_service.update_token_from_json(request.json)
        return response.json(domain_service.get_token())

    @project_endpoints.route("/domain", methods=["GET", "HEAD"])
    @platform_scoped("domain.get")
    async def get_domain(request):
        domain_yaml = DomainService(request["db_session"]).get_domain_yaml()
        if not domain_yaml:
            return error(400, "DomainNotFound", "Could not find domain.")

        return response.text(domain_yaml)

    @project_endpoints.route("/domainActions", methods=["GET", "HEAD"])
    @platform_scoped("domainActions.get")
    async def get_domain_actions(request):
        domain_actions = DomainService(request["db_session"]).get_actions_from_domain()

        if domain_actions is None:
            return error(400, "DomainNotFound", "Could not find domain.")

        # convert to list for json serialisation
        domain_actions = list(domain_actions)

        return response.json(
            domain_actions, headers={"X-Total-Count": len(domain_actions)}
        )

    @project_endpoints.route("/domain", methods=["PUT"])
    @platform_scoped("domain.update", allow_api_token=True)
    async def update_domain(request):
        domain_yaml = convert_bytes_to_string(request.body)
        try:
            updated_domain = DomainService(
                request["db_session"]
            ).validate_and_store_domain_yaml(domain_yaml, store_templates=False)
        except InvalidDomain as e:
            return error(400, "InvalidDomainError", "Could not update domain.", e)

        return response.text(updated_domain)

    @project_endpoints.route("/domainWarnings", methods=["GET", "HEAD"])
    @platform_scoped("domainWarnings.get")
    async def get_domain_warnings(request):
        domain_service = DomainService(request["db_session"])
        domain_warnings = await domain_service.get_domain_warnings()
        if domain_warnings is None:
            return error(400, "DomainNotFound", "Could not find domain.")

        return response.json(
            domain_warnings[0], headers={"X-Total-Count": domain_warnings[1]}
        )

    return project_endpoints
